#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <complex.h>

#include "vdw_splines.h"
#include "kernel.h"
#include "vdwxc.h"

#if defined HAVE_CONFIG_H
#include "config.h"
#endif

#include "vdw_core.h"

#if defined VDW_TIMING
#include "vdw_timing.h"
#endif


// Get name of functional from numeric code.
const char* vdwxc_funcname(int functional)
{
    static char* names[] = { "vdW-DF", "vdW-DF2", "vdW-DF-CX", "custom", "unknown" };
    switch(functional) {
    case FUNC_VDWDF:
        return names[0];
    case FUNC_VDWDF2:
        return names[1];
    case FUNC_VDWDFCX:
        return names[2];
    case FUNC_CUSTOM:
        return names[3];
    default:
        return names[4];
    }
}

// Get name of parallelization mode from numeric code.
const char* vdwxc_fftname(int fftcode)
{
    static char* names[] = {"fftw-serial", "fftw-mpi", "pfft", "unknown"};
    switch(fftcode) {
    case VDW_FFTW_SERIAL:
        return names[0];
    case VDW_FFTW_MPI:
        return names[1];
    case VDW_PFFT:
        return names[2];
    default:
        return names[3];
    }
}

const char* vdwxc_errstring(int errcode)
{
    static char* errstrs[] = {
        "No errorcode set.",
        "struct vdwxc_data uninitialized. Call vdwxc_initialize(...);",
        "Unkown functional code.",
        "Unit cell not set. Call vdwxc_set_unit_cell(...);",
        "Unit cell already set in call to vdwxc_set_unit_cell(...);",
        "Radial grid not set. Call vdwxc_set_radial_grid(...);",
        "Unknown error code or uninitialized vdwxc_data object."
    };
    if(errcode >= 0 && errcode <= 5)
        return errstrs[errcode];
    else
        return errstrs[6];
}

// Get the size of a vdwxc_data.
int vdwxc_get_struct_size(void)
{
    return sizeof(struct vdwxc_data_obj);
}

// Set all bytes of vdwxc_data to zero.
void vdwxc_nullify(vdwxc_data data)
{
    memset(data, 0, vdwxc_get_struct_size());
}

// Set functional type and some default values.
void vdwxc_set_defaults(vdwxc_data vdw, int functional)
{
    switch (functional) {
    case FUNC_VDWDF:
        vdw->Z_ab = -0.8491;
        break;
    case FUNC_VDWDF2:
        vdw->Z_ab = -1.887;
        break;
    case FUNC_VDWDFCX:
        vdw->Z_ab = -0.8491;
        break;
    default:
        printf("unknown vdw-df functional %d\n", functional);
        assert(0);
        break;
    }
    vdw->functional = functional;
    vdw->q_cut = 5.0;
    vdw->fft_type = VDW_FFTW_SERIAL;
    vdw->mpi_size = 1;
    vdw->initialized = VDW_INITIALIZED;
}

// Allocate a new vdwxc_data and return a pointer to it.
vdwxc_data vdwxc_new(int functional)
{
    vdwxc_data vdw = malloc(vdwxc_get_struct_size());
    vdwxc_nullify(vdw);
    vdwxc_set_defaults(vdw, functional);
    return vdw;
}

// Write unit cell to string
int vdwxc_sprint_cell(char* str, struct vdwxc_unit_cell* cell)
{
    char* ch = str;
    ch += sprintf(ch, "    Nglobal = %d x %d x %d\n", cell->Nglobal[0], cell->Nglobal[1], cell->Nglobal[2]);
    ch += sprintf(ch, "    Nlocal  = %d x %d x %d\n", cell->Nlocal[0], cell->Nlocal[1], cell->Nlocal[2]);
    ch += sprintf(ch, "    offset  = %d, %d, %d\n", cell->offset[0], cell->offset[1], cell->offset[2]);
    ch += sprintf(ch, "      %f, %f, %f\n", cell->vec[0], cell->vec[1], cell->vec[2]);
    ch += sprintf(ch, "      %f, %f, %f\n", cell->vec[3], cell->vec[4], cell->vec[5]);
    ch += sprintf(ch, "      %f, %f, %f\n", cell->vec[6], cell->vec[7], cell->vec[8]);
    ch += sprintf(ch, "    dV %e\n", cell->dV);
    int nbytes = ch - str;
    return nbytes;
}

#ifdef HAVE_MPI
int vdwxc_has_mpi(void){ return 1; }
#else
int vdwxc_has_mpi(void){ return 0; }
#endif

#ifdef HAVE_PFFT
int vdwxc_has_pfft(void){ return 1; }
#else
int vdwxc_has_pfft(void){  return 0; }
#endif

// Print representation of vdwxc_data to stdout.
void vdwxc_print(vdwxc_data data)
{
    int maxsize = 80 * 200;
    char str[maxsize];
    vdwxc_tostring(data, maxsize, str);
    // printf(str) yields a warning
    printf("%s", str);
}

// Write representation of vdwxc_data to string.  Returns number of bytes written.
// Truncates to maxsize if necessary.
int vdwxc_tostring(vdwxc_data data, int maxsize, char *str)
{
    // We build within our own buffer, then copy into user-supplied str.
    int maxlen = 200 * 80; // max 200 lines of length 80
    if(maxsize > maxlen)
        maxlen = maxsize;
    char buf[maxlen];
    char* ch = buf;
    char* start = buf;
    ch += sprintf(ch, "  === vdW-DF data at %p ===\n", data);
    ch += sprintf(ch, "  functional %s [%d]\n", vdwxc_funcname(data->functional), data->functional);
    ch += sprintf(ch, "  Z=%f :: qcut=%f\n", data->Z_ab, data->q_cut);
    ch += sprintf(ch, "  cell\n");
    ch += vdwxc_sprint_cell(ch, &data->cell);
    ch += sprintf(ch, "  icell\n");
    ch += vdwxc_sprint_cell(ch, &data->icell);
    ch += sprintf(ch, "  FFT: %s [%d]\n", vdwxc_fftname(data->fft_type), data->fft_type);
    if(data->fft_type == VDW_PFFT) {
        ch += sprintf(ch, "  pfft grid %d x %d\n", data->pfft_grid[0], data->pfft_grid[1]);
    }
#ifdef HAVE_MPI
    ch += sprintf(ch, "  MPI :: rank=%d :: size=%d :: ptr=<%p>\n",
                  data->mpi_rank, data->mpi_size, data->mpi_comm);
#endif
    ch += sprintf(ch, "  alloc Ng=%d Nglobal=%d gLDA=%d\n",
                  data->Ng, data->Nglobal, data->gLDA);
    ch += sprintf(ch, "  kLDA=%d\n", data->kLDA);
    ch += sprintf(ch, "  ptrs1 q0=<%p> work_ka=<%p>\n",
                  data->q0_g, data->work_ka);
    ch += sprintf(ch, "  ptrs2 dq0_drho_g=<%p> dq0_dsigma_g=<%p>\n",
                  data->dq0_drho_g, data->dq0_dsigma_g);
    ch += sprintf(ch, "  fftw ptrs  r2c=<%p> c2r=<%p>\n", data->plan_r2c, data->plan_c2r);
#ifdef HAVE_PFFT
    ch += sprintf(ch, "  pfft ptrs r2c=<%p> c2r=<%p>\n", data->pfft_plan_r2c, data->pfft_plan_c2r);
#endif
    ch += sprintf(ch, "  Error state %d: %s\n", data->errorcode, vdwxc_errstring(data->errorcode));
    ch += sprintf(ch, "  =============================\n");
    int len = strlen(buf);
    assert(ch - buf == len);
    assert(len < maxlen);
    assert(len < maxsize);
    strcpy(str, buf);
    return len;
}

void vdwxc_set_unit_cell(vdwxc_data data,
                         int Nx, int Ny, int Nz,
                         double C00, double C10, double C20,
                         double C01, double C11, double C21,
                         double C02, double C12, double C22)
{
    double det = (C00 * (C11 * C22 - C21 * C12) \
                  - C01 * (C10 * C22 - C12 * C20) \
                  + C02 * (C10 * C21 - C11 * C20));
    data->cell.dV = det / (Nx * Ny * Nz);
    data->cell.vec[0] = C00;
    data->cell.vec[1] = C01;
    data->cell.vec[2] = C02;
    data->cell.vec[3] = C10;
    data->cell.vec[4] = C11;
    data->cell.vec[5] = C12;
    data->cell.vec[6] = C20;
    data->cell.vec[7] = C21;
    data->cell.vec[8] = C22;

    data->cell.Nglobal[0] = Nx;
    data->cell.Nglobal[1] = Ny;
    data->cell.Nglobal[2] = Nz;

    // Calculate reciprocal cell
    double idet = 1.0 / (C00 * (C11 * C22 - C21 * C12) \
                         - C01 * (C10 * C22 - C12 * C20) \
                         + C02 * (C10 * C21 - C11 * C20)) * 2 * VDW_PI;
    data->icell.Nglobal[0] = Nx;
    data->icell.Nglobal[1] = Ny;
    data->icell.Nglobal[2] = Nz / 2 + 1;
    data->icell.dV = 1e100;
    data->icell.vec[0] =  (C11 * C22 - C21 * C12) * idet;
    data->icell.vec[1] = -(C01 * C22 - C02 * C21) * idet;
    data->icell.vec[2] =  (C01 * C12 - C02 * C11) * idet;
    data->icell.vec[3] = -(C10 * C22 - C12 * C20) * idet;
    data->icell.vec[4] =  (C00 * C22 - C02 * C20) * idet;
    data->icell.vec[5] = -(C00 * C12 - C10 * C02) * idet;
    data->icell.vec[6] =  (C10 * C21 - C20 * C11) * idet;
    data->icell.vec[7] = -(C00 * C21 - C20 * C01) * idet;
    data->icell.vec[8] =  (C00 * C11 - C10 * C01) * idet;

    data->Nglobal = data->cell.Nglobal[0]
                  * data->cell.Nglobal[1]
                  * data->cell.Nglobal[2];
}

void vdwxc_allocate_buffers(vdwxc_data data)
{
    data->Ng = data->cell.Nlocal[0] * data->cell.Nlocal[1] * data->cell.Nlocal[2];

    size_t gridsize = data->Ng * sizeof(double);

    data->q0_g = (double *) malloc(gridsize);
    data->dq0_drho_g = (double *) malloc(gridsize);
    data->dq0_dsigma_g = (double *) malloc(gridsize);
}

void vdwxc_init_serial(vdwxc_data data)
{
    data->kLDA = data->icell.Nglobal[2];
    data->gLDA = 2 * data->kLDA;
    int nktotal = data->cell.Nglobal[0] * data->cell.Nglobal[1] * data->kLDA * SPLINE_ALPHAS;
    data->work_ka = (double complex*) fftw_malloc(nktotal * sizeof(double complex));
    int i;
    for(i = 0; i < 3; i++) {
        data->cell.Nlocal[i] = data->cell.Nglobal[i];
        data->icell.Nlocal[i]= data->cell.Nglobal[i];
    }
    data->icell.Nlocal[2] = data->icell.Nlocal[2] / 2 + 1;

    data->plan_r2c = fftw_plan_many_dft_r2c(3,
                                            data->cell.Nglobal,
                                            SPLINE_ALPHAS,
                                            (double*)data->work_ka,
                                            NULL,
                                            SPLINE_ALPHAS,
                                            1,
                                            data->work_ka,
                                            NULL,
                                            SPLINE_ALPHAS,
                                            1,
                                            FFTW_ESTIMATE);
    data->plan_c2r = fftw_plan_many_dft_c2r(3,
                                            data->cell.Nglobal,
                                            SPLINE_ALPHAS,
                                            data->work_ka,
                                            NULL,
                                            SPLINE_ALPHAS,
                                            1,
                                            (double*)data->work_ka,
                                            NULL,
                                            SPLINE_ALPHAS,
                                            1,
                                            FFTW_ESTIMATE);
    assert(data->plan_r2c != NULL);
    assert(data->plan_c2r != NULL);
    vdwxc_allocate_buffers(data);
}

// Calculate potential and derivatives from F_ka.
void vdwxc_potential(vdwxc_data data, double* rho_g, double* dedn_g, double* dedsigma_g)
{
    START_TIMER(potential, "potential");
    double prefactor = 1.0 / data->Nglobal;
    double* F_ga = (double*)data->work_ka; // XXXXX
    int ia_fftw;
    int ixyz, ixyz_fftw;
    int ix, iy, iz;
    //double integrals[2] = {0.0, 0.0};
    int alpha;
    // TODO: Move this to inner loop to save Memory

    double p_a[SPLINE_ALPHAS];
    double dpdq_a[SPLINE_ALPHAS];

    for(ix = 0; ix < data->cell.Nlocal[0]; ix++) {
        for(iy = 0; iy < data->cell.Nlocal[1]; iy++) {
            for(iz = 0; iz < data->cell.Nlocal[2]; iz++) {
                ixyz = iz + data->cell.Nlocal[2] * (iy + (data->cell.Nlocal[1] * ix));
                ixyz_fftw = (iz + data->gLDA * (iy + data->cell.Nlocal[1] * ix));
                double rho_dq0_drho = rho_g[ixyz] * data->dq0_drho_g[ixyz];
                double rho_dq0_dsigma = rho_g[ixyz] * data->dq0_dsigma_g[ixyz];
                assert(ixyz < data->Ng);
                dedn_g[ixyz] = 0.0; // zero or not?  From perspective of calling code...
                dedsigma_g[ixyz] = 0.0; // zero or not?
                vdwxc_evaluate_palpha_splines(1.0, data->q0_g[ixyz], p_a);
                vdwxc_evaluate_palpha_splines_derivative(data->q0_g[ixyz], dpdq_a);
                for(alpha=0; alpha<SPLINE_ALPHAS; alpha++) {
                    ia_fftw = ixyz_fftw * SPLINE_ALPHAS + alpha;
                    dedn_g[ixyz] += prefactor * F_ga[ia_fftw] *
                        (p_a[alpha] + dpdq_a[alpha] * rho_dq0_drho);
                    dedsigma_g[ixyz] += prefactor * F_ga[ia_fftw] *
                        dpdq_a[alpha] * rho_dq0_dsigma;
                }
                // Here we can evaluate energy density as well:
                //
                //   epsilon(r) = (1/2) sum[alpha] n(r) p[alpha](r) F[alpha](r)
                //
                // n(r) times p[alpha](r) is actually theta[alpha](r),
                // which we do not have because we overwrote it.
                // But I don't know if anyone actually needs epsilon(r) for anything.

                //integrals[0] += dedn_g[ixyz] * rho_g[ixyz];
                //integrals[1] += dedsigma_g[ixyz] * rho_g[ixyz];
            }
        }
    }
    //integrals[0] *= data->cell.dV;
    //integrals[1] *= data->cell.dV;
    END_TIMER(potential, "potential");
    // store integrals somehow for checking/debugging?
}

void vdwxc_calculate_q0(int N, double Z_ab, double q_cut,
                        double* rho, double* sigma,
                        double* q0, double* dq0_drho,
                        double* dq0_dsigma)
{
    int i;
    for (i=0; i<N; i++) {
        double mrho = *rho;
        if (*rho < 1e-10) {
            mrho = 1e-10;
            //if (*rho<0)
            //   printf("Warning, negative density %.15f\n", *rho);
        }

        // kF = (3.0 * np.pi**2 * total_rho)**(1.0 / 3.0)
        double rho1_root3 = pow(mrho, 1.0 / 3.0);
        double kF = VDW_KF_PREFACTOR * rho1_root3;

        // r_s = (3.0 / (4.0 * np.pi * total_rho))**(1.0 / 3.0)
        double rs = VDW_RS_PREFACTOR / rho1_root3;
        double rs_sqr = rs*rs;
        double rs_root2 = sqrt(rs);
        double rho_sqr = mrho*mrho;

        double gradient_correction = -Z_ab / (36.0 * kF * rho_sqr) * (*sigma);
        double LDA_1 = 8.0 * VDW_PI / 3.0 * (VDW_LDA_A * (1.0 + VDW_LDA_a1 * rs));
        double LDA_2 = 2.0 * VDW_LDA_A * (VDW_LDA_b1 * rs_root2 + VDW_LDA_b2 * rs + \
                                          VDW_LDA_b3 * rs * rs_root2 + VDW_LDA_b4 * rs_sqr);

        double q = kF + LDA_1 * log(1.0 + 1.0 / LDA_2) + gradient_correction;
        int m;
        double exponent = 0.0;
        double dq0_dq = 0.0;
        double qscaled0 = q / q_cut;
        double qscaled = 1.0;

        for (m=1; m<=12; m++) {
            dq0_dq += qscaled;
            qscaled *= qscaled0;
            exponent += qscaled / m;
        }

        //printf("exponent %.15f", exponent);
        //printf("dq0_dq %.15f", dq0_dq);

        //mask = total_rho >= epsr
        //q0[mask] = (q_cut * (1.0 - np.exp(-exponent)))[mask]
        //dq0_dq *= np.exp(-exponent)
        // Ignore mask for now: DOUBLE CHECK THIS XXX
        dq0_dq *= exp(-exponent);
        *q0 = q_cut * (1.0 - exp(-exponent));

        *dq0_drho = dq0_dq * (kF / 3.0 - 7.0 / 3.0 * gradient_correction
                              - 8.0 * VDW_PI / 9.0 * VDW_LDA_A * VDW_LDA_a1 * rs
                              * log(1.0 + 1.0 / LDA_2)
                              + LDA_1 / (LDA_2 * (1.0 + LDA_2))
                              * (2.0 * VDW_LDA_A * (VDW_LDA_b1 / 6.0 * rs_root2
                                      + VDW_LDA_b2 / 3.0 * rs
                                      + VDW_LDA_b3 / 2.0 * rs * rs_root2
                                      + 2.0 * VDW_LDA_b4 / 3.0 * rs_sqr)));

        *dq0_drho /= mrho;
        *dq0_dsigma = -(dq0_dq) * 2.0 * Z_ab / (36.0 * kF * mrho);
        *dq0_dsigma /= 2.0 * mrho;

        rho++;
        sigma++;
        q0++;
        dq0_drho++;
        dq0_dsigma++;
    }
}

void vdwxc_calculate_thetas(int nx, int ny, int nz, int leading_dim,
                            double* rho, double* q0, double* output)
{
    // We have to translate between the contiguous array rho
    // and the padded array output.
    int ix, iy, iz;
    int ixyz_fftw;
    for(ix = 0; ix < nx; ix++) {
        for(iy = 0; iy < ny; iy++) {
            for(iz = 0; iz < nz; iz++) {
                //ijk = iz + nz * (iy + (ny * ix));
                ixyz_fftw = (iz + leading_dim * (iy + ny * ix));
                vdwxc_evaluate_palpha_splines(*rho, *q0, output + ixyz_fftw * SPLINE_ALPHAS);
                rho++;
                q0++;
            }
        }
    }
}

// Calculate q0 and theta from density and gradient.
void vdwxc_q0_and_theta(vdwxc_data data, double* rho_g, double* sigma_g)
{
    if (data->initialized != VDW_INITIALIZED)
    {
        data->errorcode = VDW_ERROR_UNINITIALIZED;
        vdwxc_print_error(data);
        return;
    }

    START_TIMER(q0, "q0");
    vdwxc_calculate_q0(data->Ng, data->Z_ab, data->q_cut,
                       rho_g, sigma_g, data->q0_g,
                       data->dq0_drho_g, data->dq0_dsigma_g);
    END_TIMER(q0, "q0");
    START_TIMER(thetas, "thetas");
    vdwxc_calculate_thetas(data->cell.Nlocal[0],
                           data->cell.Nlocal[1],
                           data->cell.Nlocal[2],
                           data->gLDA,
                           rho_g, data->q0_g,
                           (double*)data->work_ka);
    END_TIMER(thetas, "thetas");
}


#define MODE_SERIAL
#define METH(name) name##_serial
#include "vdw_include.c"
#undef NAME
#undef METH
#undef MODE_SERIAL

#define MODE_MPI
#define METH(name) name##_mpi
#include "vdw_include.c"
#undef NAME
#undef METH
#undef MODE_MPI

#define MODE_PFFT
#define METH(name) name##_pfft
#include "vdw_include.c"
#undef NAME
#undef METH
#undef MODE_PFFT


// Calculate van der Waals energy plus energy/density derivative and
// energy/gradient-density derivative from density and gradient.
double vdwxc_calculate(vdwxc_data data,
                       double* rho_g, double* sigma_g,
                       double* dedn_g, double* dedsigma_g)
{
    if (data->initialized != VDW_INITIALIZED)
    {
        data->errorcode = VDW_ERROR_UNINITIALIZED;
        vdwxc_print_error(data);
        return 0.0;
    }
    if (data->cell.Nglobal[0] * data->cell.Nglobal[1] * data->cell.Nglobal[2] <= 0)
    {
        data->errorcode = VDW_ERROR_UNIT_CELL_NOT_SET;
        vdwxc_print_error(data);
        return 0.0;
    }

    switch(data->fft_type) {
    case VDW_FFTW_SERIAL:
        return vdwxc_calculate_serial(data, rho_g, sigma_g, dedn_g, dedsigma_g);
    case VDW_FFTW_MPI:
        return vdwxc_calculate_mpi(data, rho_g, sigma_g, dedn_g, dedsigma_g);
    case VDW_PFFT:
        return vdwxc_calculate_pfft(data, rho_g, sigma_g, dedn_g, dedsigma_g);
    default:
        assert(0);
        return 0.0;
    }
}

void vdwxc_print_error(vdwxc_data data)
{
    printf("%s (errcode=%d)\n", vdwxc_errstring(data->errorcode), data->errorcode);
    if(data->errorcode != VDW_ERROR_NOERROR) {
        assert(0);
    }
    /*if (data->errorcode == VDW_ERROR_NOERROR) {
       printf("No errorcode set.\n");
       return;
    }

    printf("libvdwxc error:");
    switch (data->errorcode)
    {
       case VDW_ERROR_UNINITIALIZED:
           printf("struct vdwxc_data uninitialized. Call vdwxc_initialize(...);");
           break;
       case VDW_ERROR_UNKNOWN_FUNCTIONAL:
           printf("Unkown functional code.");
           break;
       case VDW_ERROR_UNIT_CELL_NOT_SET:
           printf("Unit cell not set. Call vdwxc_set_unit_cell(...);");
           break;
       case VDW_ERROR_UNIT_CELL_ALREADY_SET:
           printf("Unit cell already set in call to vdwxc_set_unit_cell(...);");
           break;
       case VDW_ERROR_RADIAL_GRID_NOT_SET:
           printf("Radial grid not set. Call vdwxc_set_radial_grid(...);");
           break;
       default:
           printf("Unknown error code or uninitialized vdwxc_data object.");
    }
    printf("\n");
    assert(0);
    */
}


void vdwxc_fatal(char* msg)
{
    printf("Fatal: %s", msg);
    assert(0); // TODO: Do things well
}

void vdwxc_finalize(vdwxc_data* vdw)
{
    vdwxc_data data = vdw[0];
#ifdef HAVE_PFFT
    if(data->fft_type == VDW_PFFT) {
        if (data->pfft_plan_r2c != NULL) {
            pfft_destroy_plan(data->pfft_plan_r2c);
        }
        if (data->pfft_plan_c2r != NULL) {
            pfft_destroy_plan(data->pfft_plan_c2r);
        }
        if (data->mpi_cart_comm != NULL) {
            MPI_Comm_free(&data->mpi_cart_comm);
        }
        if (data->work_ka != NULL) {
            pfft_free(data->work_ka);
            // Set to null so we won't call fftw_free on it in a moment
            data->work_ka = NULL;
        }
    }
#else
    assert(data->fft_type != VDW_PFFT);
#endif
    if (data->plan_r2c != NULL) {
        fftw_destroy_plan(data->plan_r2c);
    }
    if (data->plan_c2r != NULL) {
        fftw_destroy_plan(data->plan_c2r);
    }
    //if (data->radial_plan_r2c != NULL) {
    //    fftw_destroy_plan(data->radial_plan_r2c);
    //    data->radial_plan_r2c = NULL;
    //}
    //if (data->radial_plan_c2r != NULL) {
    //    fftw_destroy_plan(data->radial_plan_c2r);
    //    data->radial_plan_c2r = NULL;
    //}
    if (data->work_ka != NULL) {
        fftw_free(data->work_ka);
    }
    if (data->q0_g != NULL) {
        free(data->q0_g);
    }
    if (data->dq0_drho_g != NULL) {
        free(data->dq0_drho_g);
    }
    if (data->dq0_dsigma_g != NULL) {
        free(data->dq0_dsigma_g);
    }
    vdwxc_nullify(data);
    free(data);
    vdw[0] = NULL;
}
