#ifndef VDW_CORE_H
#define VDW_CORE_H

// libvdwxc can be compiled in three ways:
//  * libvdwxc + fftw
//  * libvdwxc + fftw + mpi + fftw-mpi
//  * libvdwxc + fftw + mpi + fftw-mpi + pfft

#include <complex.h>

#define VDW_LDA_A 0.031091
#define VDW_LDA_a1 0.2137
#define VDW_LDA_b1 7.5957
#define VDW_LDA_b2 3.5876
#define VDW_LDA_b3 1.6382
#define VDW_LDA_b4 0.49294

#define VDW_KF_PREFACTOR 3.0936677262801355 // (3.0 * pi**2)**(1.0 / 3.0)
#define VDW_RS_PREFACTOR 0.6203504908994001 // (3.0 / (4.0 * pi))**(1.0 / 3.0)

#define VDW_ERROR_NOERROR 0
#define VDW_ERROR_UNINITIALIZED 1
#define VDW_ERROR_UNKNOWN_FUNCTIONAL 2
#define VDW_ERROR_UNIT_CELL_NOT_SET 3
#define VDW_ERROR_UNIT_CELL_ALREADY_SET 4
#define VDW_ERROR_RADIAL_GRID_NOT_SET 5

#define FUNC_VDWDF 1
#define FUNC_VDWDF2 2
#define FUNC_VDWDFCX 3
#define FUNC_CUSTOM -1

#define VDW_FFTW_SERIAL 0
#define VDW_FFTW_MPI 1
#define VDW_PFFT 2

#define VDW_INITIALIZED (unsigned int) 0x76647778 // vdwx
#define VDW_UNINITIALIZED 0

typedef struct vdwxc_data_obj *vdwxc_data;

// High level functions
// There are the actual functions which will be called by DFT codes
vdwxc_data vdwxc_new(int functional);

void vdwxc_print(vdwxc_data data);
int vdwxc_tostring(vdwxc_data data, int maxsize, char *str);
void vdwxc_print_error(vdwxc_data data);

void vdwxc_set_unit_cell(vdwxc_data data,
                         int Nx, int Ny, int Nz,
                         double C00, double C10, double C20,
                         double C01, double C11, double C21,
                         double C02, double C12, double C22);

// Set the radial grid for vdW-DF calculations.
void vdwxc_df_set_radial_grid(vdwxc_data data, double h, int Nr);

// These are functions because it seems more portable than constants
// For example they can be called through Python ctypes and such.
int vdwxc_has_mpi(void);
int vdwxc_has_pfft(void);

void vdwxc_init_serial(vdwxc_data data);
double vdwxc_calculate(vdwxc_data data,
                       double* rho_i, double* sigma_i,
                       double* dedn_i, double* dedsigma_i);

double vdwxc_calculate_radial(vdwxc_data data, int N, double dr, double* rho_i,
                            double* sigma_i, double* dedn_i, double* dedsigma_i);


void vdwxc_finalize(vdwxc_data* data);

// Functions for more or less internal use.
void vdwxc_allocate_buffers(vdwxc_data data);


/*
Input:
  int N : number of grid points
  double Z_ab
  double q_cut
  double* rho : density (in electrons / Bohr^3)
  double* sigma : |\nabla \rho|^2

Output:
  double* q0 : q0
  double* dq0_drho : Derivative of q0 wrt. rho
  double* dq0_dsigma : Derivative of q0 wrt. gradient of rho
*/
void vdwxc_calculate_q0(int N, double Z_ab, double q_cut,
                        double* rho, double* sigma,
                        double* q0, double* dq0_drho,
                        double* dq0_dsigma);
void vdwxc_add_cx_exchange(vdwxc_data data, double* rho_g, double* sigma_g,
                           double* eps_g, double* dedn_g, double* dedsigma_g);
void vdwxc_cx_semilocal_exchange(int ng, double* rho_g, double* sigma_g,
                                 double* eps_g, double* dedn_g,
                                 double* dedsigma_g);
void vdwxc_q0_and_theta(vdwxc_data data, double* rho_g, double* sigma_g);
void vdwxc_fft_r2c(vdwxc_data data);
void vdwxc_fft_c2r(vdwxc_data data);
double vdwxc_convolution(vdwxc_data data);
void vdwxc_potential(vdwxc_data data, double* rho_g, double* dedn_g, double* dedsigma_g);
void vdwxc_fatal(char* msg);

#endif
